Attribute VB_Name = "modSimplx"
'======================================================
' Descrizione.....: Routines di Programmazione Lineare.
' Nome dei Files..: modSimplx.bas
' Data............: 3/7/1999
' Revisione.......: 26/1/2001
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 sotto Windows NT.
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
'======================================================
'
'   Tradotto dalle Subroutines FORTRAN di:
'    Numerical Recipes in FORTRAN 77
'    Chapter 10. Minimization or Maximization of Functions.
'    The Art of Scientific Computing
'    Second Edition Volume 1 of
'    FORTRAN Numerical Recipes
'    di William H. Press ed altri
'
Option Explicit

Public Sub Simplx(a#(), ByVal m1&, ByVal m2&, ByVal m3& _
    , ICASE&, IZROV&(), IPOSV&(), Optional ByVal eps# = 0.000001)
'
'   Simplex method for linear programming. Input parameters a(), m1, m2, and m3,
'   and output parameters a(), ICASE, IZROV(), and IPOSV() are described in Dati.
'   Parameters:  EPS is the absolute precision, which should be adjusted to
'   the scale of your variables.
'
    Dim I&, IP&, IX&, K&, KH&, KP&, M&, N&, NL1&
    Dim bmax#, q1#
'
    M = UBound(a, 1) - 2
    N = UBound(a, 2) - 1
    ReDim IPOSV(1 To M), IZROV(1 To N)
    ReDim L1&(1 To N), L3&(1 To M)
'
    If (M <> m1 + m2 + m3) Then
        MsgBox "Bad input constraint counts in Simplx.", vbCritical, " Simplx"
        ICASE = -2
        Exit Sub
    End If
'
    NL1 = N
    For K = 1 To N
        L1(K) = K       ' Initialize index list of columns admissible for exchange.
        IZROV(K) = K    ' Initially make all variables right-hand.
    Next K
    For I = 1 To M
        If (a(I + 1, 1) < 0#) Then  ' Constants bi must be non-negative.
            MsgBox "Bad input tableau in Simplx.", vbCritical, " Simplx"
            ICASE = -2
            Exit Sub
        End If
'
        IPOSV(I) = N + I
        ' Initial left-hand variables.
        ' m1 type constraints are represented by having their slack variable initially
        '    left-hand, with no artificial variable.
        ' m2 type constraints have their slack variable initially left-hand, with a
        '    minus sign, and their artificial variable handled implicitly during their
        '    first exchange.
        ' m3 type constraints have their artificial variable initially left-hand.
    Next I
    If (m2 + m3 = 0) Then GoTo 30   ' The origin is a feasible starting solution.
                                    ' Go to phase two.
'
    For I = 1 To m2                 ' Initialize list of m2 constraints whose slack
        L3(I) = 1                   ' variables have never been exchanged out of the
    Next I                          ' initial basis.
'
    For K = 1 To N + 1              ' Compute the auxiliary objective function.
        q1 = 0#
        For I = m1 + 1 To M
            q1 = q1 + a(I + 1, K)
        Next I
        a(M + 2, K) = -q1
    Next K
'
10  Simp1 a(), M + 1, L1(), NL1, 0, KP, bmax        ' Find max. coeff. of auxiliary
                                                    ' objective fn.
    If (bmax <= eps) And (a(M + 2, 1) < -eps) Then
        ICASE = -1                                  ' Auxiliary objective function is
        Exit Sub                                    ' still negative and can't be im-
                                                    ' proved, hence no feasible solution
                                                    ' exists.
'
    ElseIf (bmax <= eps) And (a(M + 2, 1) <= eps) Then
    ' Auxiliary objective function is zero and can't be improved; we have a feasible
    ' starting vector. Clean out the artificial variables corresponding to any
    ' remaining equality constraints by GoTo 1's and then move on to phase two by
    ' GoTo 30.
        For IP = m1 + m2 + 1 To M
            If IPOSV(IP) = IP + N Then              ' Found an artificial variable for
                                                    ' an equality constraint.
                Simp1 a(), IP, L1(), NL1, 1, KP, bmax
                If (bmax > eps) Then GoTo 1         ' Exchange with column corresponding
                                                    ' to maximum pivot element in row.
            End If
        Next IP
        For I = m1 + 1 To m1 + m2                   ' Change sign of row for any m2
                                                    ' constraints still present from
                                                    ' the initial basis.
            If L3(I - m1) = 1 Then
                For K = 1 To N + 1
                    a(I + 1, K) = -a(I + 1, K)
                Next K
            End If
        Next I
        GoTo 30                                     ' Go to phase two.
    End If
'
    Simp2 a(), M, N, IP, KP, eps                    ' Locate a pivot element (phase one).
    If IP = 0 Then                                  ' Maximum of auxiliary objective
        ICASE = -1                                  ' function is unbounded, so no
        Exit Sub                                    ' feasible solution exists.
    End If
'
1   Simp3 a(), M + 1, N, IP, KP
    ' Exchange a left- and a right-hand variable (phase one), then update lists.
    If IPOSV(IP) >= (N + m1 + m2 + 1) Then          ' Exchanged out an artificial variable
                                                    ' for an equality constraint. Make
                                                    ' sure it stays out by removing it
                                                    ' from the l1 list.
        For K = 1 To NL1
            If L1(K) = KP Then GoTo 2
        Next K
'
2       NL1 = NL1 - 1
        For IX = K To NL1
            L1(IX) = L1(IX + 1)
        Next IX
    Else
        KH = IPOSV(IP) - m1 - N
        If KH >= 1 Then                             ' Exchanged out an m2 type constraint.
            If L3(KH) <> 0 Then                     ' If it's the first time, correct
                                                    ' the pivot column for the minus
                                                    ' sign and the implicit artificial
                                                    ' variable.
                L3(KH) = 0
                a(M + 2, KP + 1) = a(M + 2, KP + 1) + 1#
                For I = 1 To M + 2
                    a(I, KP + 1) = -a(I, KP + 1)
                Next I
            End If
        End If
    End If
    IX = IZROV(KP)                                  ' Update lists of left- and
                                                    ' right-hand variables.
    IZROV(KP) = IPOSV(IP)
    IPOSV(IP) = IX
    GoTo 10                                         ' Still in phase one, go back to 10.
    ' End of phase one code for finding an initial feasible solution.
'
    ' Now, in phase two, optimize it.
30  Simp1 a(), 0, L1(), NL1, 0, KP, bmax            ' Test the z-row for doneness.
    If bmax <= eps Then                             ' Done. Solution found.
        ICASE = 0                                   ' Return with the good news.
        Exit Sub
    End If
'
    Simp2 a(), M, N, IP, KP, eps                    ' Locate a pivot element (phase two).
    If IP = 0 Then                                  ' Objective function is unbounded.
        ICASE = 1                                   ' Report and return.
        Exit Sub
    End If
'
    Simp3 a(), M, N, IP, KP                         ' Exchange a left- and a right-hand
                                                    ' variable (phase two),
    IX = IZROV(KP)                                  ' update lists of left- and
                                                    ' right-hand variables,
    IZROV(KP) = IPOSV(IP)
    IPOSV(IP) = IX
    GoTo 30                                         ' and return for another iteration.
'
'
'
End Sub

Private Sub Simp1(a#(), ByVal MM&, LL&(), ByVal NLL&, ByVal IABF&, KP&, bmax#)
'
'   Determines the maximum of those elements whose index is contained in the
'   supplied list LL(), either with or without taking the absolute value, as
'   flagged by IABF.
'
    Dim K&, test#
'
    If NLL <= 0 Then    ' No eligible columns.
        bmax = 0#
    Else
        KP = LL(1)
        bmax = a(MM + 1, KP + 1)
        For K = 2 To NLL
            If IABF = 0 Then
                test = a(MM + 1, LL(K) + 1) - bmax
            Else
                test = Abs(a(MM + 1, LL(K) + 1)) - Abs(bmax)
            End If
            If test > 0# Then
                bmax = a(MM + 1, LL(K) + 1)
                KP = LL(K)
            End If
        Next K
    End If
'
'
'
End Sub

Private Sub Simp2(a#(), ByVal M&, ByVal N&, IP&, ByVal KP&, ByVal eps#)
'
'   Locate a pivot element, taking degeneracy into account.
'
    Dim I&, K&
    Dim q#, q0#, q1#, qp#
    'Const eps = 0.000001
'
    IP = 0
    For I = 1 To M
        If (a(I + 1, KP + 1) < -eps) Then GoTo 1
    Next I
    MsgBox "No possible pivots.", vbInformation, " Simplx"
    Exit Sub    ' Return with message.
'
1   q1 = -a(I + 1, 1) / a(I + 1, KP + 1)
    IP = I
    For I = IP + 1 To M
        If (a(I + 1, KP + 1) < -eps) Then
            q = -a(I + 1, 1) / a(I + 1, KP + 1)
            If q < q1 Then
                IP = I
                q1 = q
            ElseIf q = q1 Then      ' We have a degeneracy.
                For K = 1 To N
                    qp = -a(IP + 1, K + 1) / a(IP + 1, KP + 1)
                    q0 = -a(I + 1, K + 1) / a(I + 1, KP + 1)
                    If (q0 <> qp) Then GoTo 2
                Next K
2               If (q0 < qp) Then IP = I
            End If
        End If
    Next I
'
'
'
End Sub

Private Sub Simp3(a#(), ByVal I1&, ByVal K1&, ByVal IP&, ByVal KP&)
'
'   Matrix operations to exchange a left-hand and right-hand variable (see text).
'
    Dim II&, KK&
    Dim piv#
'
    piv = 1# / a(IP + 1, KP + 1)
    For II = 1 To I1 + 1
        If II - 1 <> IP Then
            a(II, KP + 1) = a(II, KP + 1) * piv
            For KK = 1 To K1 + 1
                If KK - 1 <> KP Then
                    a(II, KK) = a(II, KK) - a(IP + 1, KK) * a(II, KP + 1)
                End If
            Next KK
        End If
    Next II
'
    For KK = 1 To K1 + 1
        If (KK - 1 <> KP) Then a(IP + 1, KK) = -a(IP + 1, KK) * piv
    Next KK
    a(IP + 1, KP + 1) = piv
'
'
'
End Sub

Private Sub Dati()
'
'
'   Forma del problema di programmazione lineare.
'
'   Massimizzare la:
'    z = a01 * x1 + a02 * x2 + ... + a0n * xn     con x1 >= 0, x2 >= 0, ..., xn >= 0,
'
'   soggetta ai vincoli:
'    m1 disequazioni di tipo 1:
'    ai1 * x1 + ai2 * x2 + ... + ain * xn <= bi   con i = 1, ..., m1
'
'    m2 disequazioni di tipo 2:
'    aj1 * x1 + aj2 * x2 + ... + ajn * xn >= bj   con j = m1 + 1, ..., m1 + m2
'
'    m3 equazioni di tipo 3:
'    ak1 * x1 + ak2 * x2 + ... + akn * xn  = bk   con k = m1 + m2 + 1, ..., m1 + m2 + m3
'
'    e con tutti i b >= 0.
'
'   I dati del problema, passati a Simplx a(), m1, m2, m3, ...,
'   devono essere organizzati come:
'
'    N =  N. di incognite.
'
'    m1 = N. di disequazioni di tipo 1:
'    m2 = N. di disequazioni di tipo 2:
'    m3 = N. di equazioni di tipo 3:
'    M  =  m1 + m2 + m3
'
'    MP = M + 2: NP = N + 1
'    ReDim a#(1 To MP, 1 To NP), IPOSV&(1 To M), IZROV&(1 To N)
'
'   I valori nella matrice a() in ingresso a Simplx sono:
'
'    0. prima riga con i coefficienti dell' equazione da massimizzare organizzati come:
'       a(1, 1) = 0:   a(1, 2) = a01:   a(1, 2) = a01: ...:  a(1, N + 1) = a0n
'
'    1. m1 righe con i coefficienti delle equazioni di vincolo <= bi organizzati come:
'       a(2, 1) = bi:  a(2, 2) = -ai1:  a(2, 3) = -ai2: ...: a(2, N + 1) = -ain
'       ......
'
'    2. m2 righe con i coefficienti delle equazioni di vincolo >= bj organizzati come:
'       a(f, 1) = bj:  a(f, 2) = -aj1:  a(f, 3) = -aj2: ...: a(f, N + 1) = -ajn
'       ......
'       con f = m1 + 1
'
'    3. m3 righe con i coefficienti delle equazioni di vincolo = bk organizzati come:
'       a(g, 1) = bk:  a(g, 2) = -ak1:  a(g, 3) = -ak2: ...: a(g, N + 1) = -akn
'       ......
'       con g = m1 + m2 + 1
'
'       Alcuni, dei tre tipi di vincolo, possono mancare: in questo caso porre
'       il corrispondente m. = 0 e non scrivere le sue righe in a().
'
'   Le soluzioni del problema, ritornate da Simplx a(), ..., ICASE, IZROV(), IPOSV()
'   sono date da:
'
'    ICASE =  0 se e' stata trovata una soluzione finita.
'          = +1 se la soluzione non e' limitata.
'          = -1 se i vincoli non possono essere soddisfatti.
'          = -2 se vi sono errori nei parametri passati a Simplx.
'
'   Se ICASE = 0 si possono conoscere i valori x della soluzione con:
'
'    Redim X#(1 To N)  ' Vettore delle soluzioni.
'    For L = 1 To M
'        If IPOSV(L) <= N Then
'            X(IPOSV(L)) = a(L + 1, 1)
'        End If
'    Next L
'
'    a(1, 1) = Valore di z(X(1), X(2), ..., X(N))
'
'   Per una descrizione piu' dettagliata dell' algoritmo usato
'   e dei valori ritornati da IZROV() e IPOSV(), vedere il testo
'   citato in Declarations.
'
'
End Sub
